==> manaplus-1.0.9.4.tar.bz2: src/gui/truetypefont.cpp <==
/*
 *  The Mana Client
 *  Copyright (C) 2004-2009  The Mana World Development Team
 *  Copyright (C) 2009-2010  The Mana Developers
 *  Copyright (C) 2009  Aethyra Development Team
 *
 *  This file is part of The Mana Client.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "gui/truetypefont.h"

#include "graphics.h"
#include "log.h"

#include "resources/image.h"
#include "resources/resourcemanager.h"

#include "utils/stringutils.h"

#include <guichan/exception.hpp>

const unsigned int CACHE_SIZE = 256;

char *strBuf;

class TextChunk
{
    public:
        TextChunk(const std::string &text, const gcn::Color &color) :
            img(0), text(text), color(color)
        {
        }

        ~TextChunk()
        {
            delete img;
            img = 0;
        }

        bool operator==(const TextChunk &chunk) const
        {
            return (chunk.text == text && chunk.color == color);
        }

        void generate(TTF_Font *font)
        {
            SDL_Color sdlCol;
            sdlCol.b = static_cast<Uint8>(color.b);
            sdlCol.r = static_cast<Uint8>(color.r);
            sdlCol.g = static_cast<Uint8>(color.g);

            getSafeUtf8String(text, strBuf);
            SDL_Surface *surface = TTF_RenderUTF8_Blended(
                    font, strBuf, sdlCol);

            if (!surface)
            {
                throw "Rendering font to surface failed: " +
                    std::string(TTF_GetError());
            }

            img = Image::load(surface);

            SDL_FreeSurface(surface);
        }

        Image *img;
        std::string text;
        gcn::Color color;
};

typedef std::list<TextChunk>::iterator CacheIterator;

static int fontCounter;

TrueTypeFont::TrueTypeFont(const std::string &filename, int size, int style)
{
    ResourceManager *resman = ResourceManager::getInstance();

    if (fontCounter == 0 && TTF_Init() == -1)
    {
        throw GCN_EXCEPTION("Unable to initialize SDL_ttf: " +
            std::string(TTF_GetError()));
    }

    if (!fontCounter)
    {
        strBuf = new char[65535];
        memset(strBuf, 65535, 0);
    }

    ++fontCounter;
    mFont = TTF_OpenFont(resman->getPath(filename).c_str(), size);

    if (!mFont)
    {
        logger->log("Error finding font " + filename);
        mFont = TTF_OpenFont(resman->getPath(
            "fonts/dejavusans.ttf").c_str(), size);
        if (!mFont)
        {
            throw GCN_EXCEPTION("SDLTrueTypeFont::SDLTrueTypeFont: " +
                                std::string(TTF_GetError()));
        }
    }

    TTF_SetFontStyle(mFont, style);
}

TrueTypeFont::~TrueTypeFont()
{
    TTF_CloseFont(mFont);
    mFont = 0;
    --fontCounter;

    if (fontCounter == 0)
    {
        TTF_Quit();
        delete []strBuf;
    }
}

void TrueTypeFont::loadFont(const std::string &filename, int size, int style)
{
    ResourceManager *resman = ResourceManager::getInstance();

    if (fontCounter == 0 && TTF_Init() == -1)
    {
        logger->log("Unable to initialize SDL_ttf: " +
                    std::string(TTF_GetError()));
        return;
    }

    TTF_Font *font = TTF_OpenFont(resman->getPath(filename).c_str(), size);

    if (!font)
    {
        logger->log("SDLTrueTypeFont::SDLTrueTypeFont: " +
                    std::string(TTF_GetError()));
        return;
    }

    if (mFont)
        TTF_CloseFont(mFont);

    mFont = font;
    TTF_SetFontStyle(mFont, style);
    clear();
}

void TrueTypeFont::clear()
{
    for (unsigned short f = 0; f < (unsigned short)CACHES_NUMBER; f ++)
        mCache[static_cast<unsigned short>(f)].clear();
}

void TrueTypeFont::drawString(gcn::Graphics *graphics,
                              const std::string &text,
                              int x, int y)
{
    if (text.empty())
        return;

    Graphics *g = dynamic_cast<Graphics *>(graphics);

    if (!g)
    {
        throw "Not a valid graphics object!";
    }

    gcn::Color col = g->getColor();
    const float alpha = static_cast<float>(col.a) / 255.0f;

    /* The alpha value is ignored at string generation so avoid caching the
     * same text with different alpha values.
     */
    col.a = 255;

    TextChunk chunk(text, col);

    unsigned char chr = text[0];
    std::list<TextChunk> *cache = &mCache[chr];

    bool found = false;

    for (CacheIterator i = cache->begin(); i != cache->end(); ++i)
    {
        if (chunk == (*i))
        {
            // Raise priority: move it to front
            cache->splice(cache->begin(), *cache, i);
            found = true;
            break;
        }
    }

    // Surface not found
    if (!found)
    {
        if (cache->size() >= CACHE_SIZE)
        {
            if (logger)
            {
//                logger->log1("TrueTypeFont remove");
                cache->pop_back();
            }
        }
        cache->push_front(chunk);
        cache->front().generate(mFont);
    }

    cache->front().img->setAlpha(alpha);
    g->drawImage(cache->front().img, x, y);
}

int TrueTypeFont::getWidth(const std::string &text) const
{
    if (text.empty())
        return 0;

    unsigned char chr = text[0];
    std::list<TextChunk> *cache = &mCache[chr];

    for (CacheIterator i = cache->begin(); i != cache->end(); i++)
    {
        if (i->text == text)
        {
            // Raise priority: move it to front
            // Assumption is that TTF::draw will be called next
            cache->splice(cache->begin(), *cache, i);
            return i->img->getWidth();
        }
    }

    int w, h;
    getSafeUtf8String(text, strBuf);
    TTF_SizeUTF8(mFont, strBuf, &w, &h);
    return w;
}

int TrueTypeFont::getHeight() const
{
    return TTF_FontHeight(mFont);
}

==> manaplus-1.0.9.12.tar.bz2: src/gui/truetypefont.cpp <==
/*
 *  The Mana Client
 *  Copyright (C) 2004-2009  The Mana World Development Team
 *  Copyright (C) 2009-2010  The Mana Developers
 *  Copyright (C) 2009  Aethyra Development Team
 *
 *  This file is part of The Mana Client.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "gui/truetypefont.h"

#include "graphics.h"
#include "log.h"

#include "resources/image.h"
#include "resources/resourcemanager.h"

#include "utils/stringutils.h"

#include <guichan/exception.hpp>

const unsigned int CACHE_SIZE = 256;

char *strBuf;

class TextChunk
{
    public:
        TextChunk(const std::string &text, const gcn::Color &color) :
            img(0), text(text), color(color)
        {
        }

        ~TextChunk()
        {
            delete img;
            img = 0;
        }

        bool operator==(const TextChunk &chunk) const
        {
            return (chunk.text == text && chunk.color == color);
        }

        void generate(TTF_Font *font)
        {
            SDL_Color sdlCol;
            sdlCol.b = static_cast<Uint8>(color.b);
            sdlCol.r = static_cast<Uint8>(color.r);
            sdlCol.g = static_cast<Uint8>(color.g);

            getSafeUtf8String(text, strBuf);
            SDL_Surface *surface = TTF_RenderUTF8_Blended(
                    font, strBuf, sdlCol);

            if (!surface)
            {
                throw "Rendering font to surface failed: " +
                    std::string(TTF_GetError());
            }

            img = Image::load(surface);

            SDL_FreeSurface(surface);
        }

        Image *img;
        std::string text;
        gcn::Color color;
};

typedef std::list<TextChunk>::iterator CacheIterator;

static int fontCounter;

TrueTypeFont::TrueTypeFont(const std::string &filename, int size, int style)
{
    ResourceManager *resman = ResourceManager::getInstance();

    if (fontCounter == 0 && TTF_Init() == -1)
    {
        throw GCN_EXCEPTION("Unable to initialize SDL_ttf: " +
            std::string(TTF_GetError()));
    }

    if (!fontCounter)
    {
        strBuf = new char[65535];
        memset(strBuf, 65535, 0);
    }

    ++fontCounter;
    mFont = TTF_OpenFont(resman->getPath(filename).c_str(), size);

    if (!mFont)
    {
        logger->log("Error finding font " + filename);
        mFont = TTF_OpenFont(resman->getPath(
            "fonts/dejavusans.ttf").c_str(), size);
        if (!mFont)
        {
            throw GCN_EXCEPTION("SDLTrueTypeFont::SDLTrueTypeFont: " +
                                std::string(TTF_GetError()));
        }
    }

    TTF_SetFontStyle(mFont, style);
}

TrueTypeFont::~TrueTypeFont()
{
    TTF_CloseFont(mFont);
    mFont = 0;
    --fontCounter;

    if (fontCounter == 0)
    {
        TTF_Quit();
        delete []strBuf;
    }
}

void TrueTypeFont::loadFont(const std::string &filename, int size, int style)
{
    ResourceManager *resman = ResourceManager::getInstance();

    if (fontCounter == 0 && TTF_Init() == -1)
    {
        logger->log("Unable to initialize SDL_ttf: " +
                    std::string(TTF_GetError()));
        return;
    }

    TTF_Font *font = TTF_OpenFont(resman->getPath(filename).c_str(), size);

    if (!font)
    {
        logger->log("SDLTrueTypeFont::SDLTrueTypeFont: " +
                    std::string(TTF_GetError()));
        return;
    }

    if (mFont)
        TTF_CloseFont(mFont);

    mFont = font;
    TTF_SetFontStyle(mFont, style);
    clear();
}

void TrueTypeFont::clear()
{
    for (unsigned short f = 0; f < (unsigned short)CACHES_NUMBER; f ++)
        mCache[static_cast<unsigned short>(f)].clear();
}

void TrueTypeFont::drawString(gcn::Graphics *graphics,
                              const std::string &text,
                              int x, int y)
{
    if (text.empty())
        return;

    Graphics *g = dynamic_cast<Graphics *>(graphics);

//    if (!g)
//        throw "Not a valid graphics object!";

    gcn::Color col = g->getColor();
    const float alpha = static_cast<float>(col.a) / 255.0f;

    /* The alpha value is ignored at string generation so avoid caching the
     * same text with different alpha values.
     */
    col.a = 255;

    TextChunk chunk(text, col);

    unsigned char chr = text[0];
    std::list<TextChunk> *cache = &mCache[chr];

    bool found = false;

    for (CacheIterator i = cache->begin(); i != cache->end(); ++i)
    {
        if (chunk == (*i))
        {
            // Raise priority: move it to front
            cache->splice(cache->begin(), *cache, i);
            found = true;
            break;
        }
    }

    // Surface not found
    if (!found)
    {
        if (cache->size() >= CACHE_SIZE)
            cache->pop_back();
        cache->push_front(chunk);
        cache->front().generate(mFont);
    }

    cache->front().img->setAlpha(alpha);
    g->drawImage(cache->front().img, x, y);
}

int TrueTypeFont::getWidth(const std::string &text) const
{
    if (text.empty())
        return 0;

    unsigned char chr = text[0];
    std::list<TextChunk> *cache = &mCache[chr];

    for (CacheIterator i = cache->begin(); i != cache->end(); i++)
    {
        if (i->text == text)
        {
            // Raise priority: move it to front
            // Assumption is that TTF::draw will be called next
            cache->splice(cache->begin(), *cache, i);
            return i->img->getWidth();
        }
    }

    int w, h;
    getSafeUtf8String(text, strBuf);
    TTF_SizeUTF8(mFont, strBuf, &w, &h);
    return w;
}

int TrueTypeFont::getHeight() const
{
    return TTF_FontHeight(mFont);
}
